#=========================================================================
# BehavioralTranslatorL5_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 20, 2019
"""Test the level 5 ehavioral translator."""

from pymtl3.datatypes import Bits16
from pymtl3.dsl import Component, OutPort
from pymtl3.passes.rtlir.util.test_utility import do_test

from ..BehavioralTranslatorL5 import BehavioralTranslatorL5
from .TestBehavioralTranslator import mk_TestBehavioralTranslator


def local_do_test( m ):
  tr = mk_TestBehavioralTranslator(BehavioralTranslatorL5)(m)
  tr.clear( m )
  tr.translate_behavioral( m )
  for component, _ref_upblk_repr in m._ref_upblk_repr.items():
    upblk_src = tr.behavioral.upblk_srcs[component]
    assert upblk_src == _ref_upblk_repr
  for component, _ref_freevar_repr in m._ref_freevar_repr.items():
    decl_freevars = tr.behavioral.decl_freevars[component]
    assert decl_freevars == _ref_freevar_repr
  for component, _ref_tmpvar_repr in m._ref_tmpvar_repr.items():
    decl_tmpvars = tr.behavioral.decl_tmpvars[component]
    assert decl_tmpvars == _ref_tmpvar_repr

def test_multi_components_tmpvars( do_test ):
  class B( Component ):
    def construct( s ):
      s.out = OutPort( Bits16 )
      @s.update
      def upblk():
        u = Bits16(0)
        s.out = u
  class A( Component ):
    def construct( s ):
      s.b = B()
      s.out = OutPort( Bits16 )
      @s.update
      def upblk():
        u = s.b.out
        s.out = u
  a = A()
  a.elaborate()
  a._ref_upblk_repr = { a : \
"""\
upblk_srcs:
  upblk_src: upblk
""", a.b : \
"""\
upblk_srcs:
  upblk_src: upblk
""" }
  a._ref_freevar_repr = { a : "freevars:\n", a.b : "freevars:\n" }
  a._ref_tmpvar_repr = { a : \
"""\
tmpvars:
  tmpvar: u in upblk of Vector16
""", a.b : \
"""\
tmpvars:
  tmpvar: u in upblk of Vector16
""" }
  do_test( a )

def test_multi_components_freevars( do_test ):
  STATE_IDLE = 0
  STATE_WORK = 1
  class B( Component ):
    def construct( s ):
      s.out = OutPort( Bits16 )
      @s.update
      def upblk():
        if 1:
          s.out = Bits16(STATE_IDLE)
        else:
          s.out = Bits16(STATE_WORK)
  class A( Component ):
    def construct( s ):
      s.b = B()
      s.out = OutPort( Bits16 )
      @s.update
      def upblk():
        if 1:
          s.out = s.b.out
        else:
          s.out = Bits16(STATE_IDLE)
  a = A()
  a.elaborate()
  a._ref_upblk_repr = { a : \
"""\
upblk_srcs:
  upblk_src: upblk
""", a.b : \
"""\
upblk_srcs:
  upblk_src: upblk
""" }
  a._ref_freevar_repr = { a.b : \
"""\
freevars:
  freevar: STATE_IDLE
  freevar: STATE_WORK
""", a : \
"""\
freevars:
  freevar: STATE_IDLE
""" }
  a._ref_tmpvar_repr = {}
  do_test( a )



"""
========================================================================
GetGiveIfc.py
========================================================================
RTL implementation of en/rdy micro-protocol.

Author: Yanghui Ou
  Date: Mar 19, 2019
"""
import greenlet

from pymtl3 import *
from pymtl3.stdlib.connects import connect_pairs

from .ifcs_utils import enrdy_to_str

#-------------------------------------------------------------------------
# GetIfcRTL
#-------------------------------------------------------------------------

class GetIfcRTL( Interface ):

  def construct( s, Type ):
    s.MsgType = Type

    s.msg = InPort ( Type )
    s.en  = OutPort( int if Type is int else Bits1 )
    s.rdy = InPort ( int if Type is int else Bits1 )

  def line_trace( s ):
    try:
      trace_len = s.trace_len
    except AttributeError:
      s.trace_len = len( "{}".format( s.MsgType() ) )
      trace_len = s.trace_len

    return enrdy_to_str( s.msg, s.en, s.rdy, trace_len )

  def __str__( s ):
    return s.line_trace()

#-------------------------------------------------------------------------
# GiveIfcRTL
#-------------------------------------------------------------------------

class GiveIfcRTL( Interface ):

  def construct( s, Type ):
    s.MsgType = Type

    s.msg = OutPort( Type )
    s.en  = InPort ( int if Type is int else Bits1 )
    s.rdy = OutPort( int if Type is int else Bits1 )

  def line_trace( s ):
    try:
      trace_len = s.trace_len
    except AttributeError:
      s.trace_len = len( "{}".format( s.MsgType() ) )
      trace_len = s.trace_len

    return enrdy_to_str( s.msg, s.en, s.rdy, trace_len )

  def __str__( s ):
    return s.line_trace()

class GetIfcFL( Interface ):

  def construct( s ):
    s.method = CallerPort()

  def __call__( s, *args, **kwargs ):
    return s.method( *args, **kwargs )

  def line_trace( s ):
    return ''

  def __str__( s ):
    return s.line_trace()

  def connect( s, other, parent ):

    # We are doing SendCL (other) -> [ RecvCL -> GiveIfcFL ] -> GetIfcFL (s)
    # SendCL is a caller interface
    if isinstance( other, NonBlockingCallerIfc ):
      m = RecvCL2GiveFL()

      if hasattr( parent, "RecvCL2GiveFL_count" ):
        count = parent.RecvCL2GiveFL_count
        setattr( parent, "RecvCL2GiveFL_" + str( count ), m )
      else:
        parent.RecvCL2GiveFL_count = 0
        parent.RecvCL2GiveFL_0 = m

      connect_pairs(
        other,  m.recv,
        m.give, s
      )
      parent.RecvCL2GiveFL_count += 1
      return True

    return False

class GiveIfcFL( Interface ):

  def construct( s, method ):
    s.method = CalleePort( method=method )

  def line_trace( s ):
    return ''

  def __call__( s, *args, **kwargs ):
    return s.method( *args, **kwargs )

  def __str__( s ):
    return s.line_trace()

#-------------------------------------------------------------------------
# RecvCL2SendRTL
#-------------------------------------------------------------------------

class RecvCL2GiveFL( Component ):

  @blocking
  def give( s ):
    while s.entry is None:
      greenlet.getcurrent().parent.switch(0)
    ret = s.entry
    s.entry = None
    return ret

  @non_blocking( lambda s : s.entry is None )
  def recv( s, msg ):
    s.entry = msg

  def construct( s ):

    # Interface

    s.give = GiveIfcFL( s.give )

    s.entry = None

    s.add_constraints( M( s.recv ) == M( s.give ) )

  def line_trace( s ):
    return "{}(){}".format( s.recv, s.give )



#=========================================================================
# SVTranslator_L2_cases_test.py
#=========================================================================
"""Test the SystemVerilog translator."""

import pytest

from pymtl3.datatypes import Bits1, Bits16, Bits32, Bits96, bitstruct, concat
from pymtl3.dsl import Component, InPort, OutPort, Wire, connect
from pymtl3.passes.rtlir.util.test_utility import do_test

from ..structural.test.SVStructuralTranslatorL1_test import check_eq
from ..SVTranslator import SVTranslator


def local_do_test( m ):
  m.elaborate()
  tr = SVTranslator( m )
  tr.translate( m )
  check_eq( tr.hierarchy.src, m._ref_src )

#-------------------------------------------------------------------------
# Behavioral
#-------------------------------------------------------------------------

def test_if( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_1 = InPort( Bits32 )
      s.in_2 = InPort( Bits32 )
      s.out = OutPort( Bits32 )
      @s.update
      def upblk():
        if Bits1(1):
          s.out = s.in_1
        else:
          s.out = s.in_2
  a = A()
  a._ref_src = \
"""
module A
(
  input logic [0:0] clk,
  input logic [31:0] in_1,
  input logic [31:0] in_2,
  output logic [31:0] out,
  input logic [0:0] reset
);

  // @s.update
  // def upblk():
  //   if Bits1(1):
  //     s.out = s.in_1
  //   else:
  //     s.out = s.in_2

  always_comb begin : upblk
    if ( 1'd1 ) begin
      out = in_1;
    end
    else
      out = in_2;
  end

endmodule
"""
  a._ref_src_yosys = a._ref_src
  do_test( a )

def test_if_dangling_else_inner( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_1 = InPort( Bits32 )
      s.in_2 = InPort( Bits32 )
      s.out = OutPort( Bits32 )
      @s.update
      def upblk():
        if Bits1(1):
          if Bits1(0):
            s.out = s.in_1
          else:
            s.out = s.in_2
  a = A()
  a._ref_src = \
"""
module A
(
  input logic [0:0] clk,
  input logic [31:0] in_1,
  input logic [31:0] in_2,
  output logic [31:0] out,
  input logic [0:0] reset
);

  // @s.update
  // def upblk():
  //   if Bits1(1):
  //     if Bits1(0):
  //       s.out = s.in_1
  //     else:
  //       s.out = s.in_2

  always_comb begin : upblk
    if ( 1'd1 ) begin
      if ( 1'd0 ) begin
        out = in_1;
      end
      else
        out = in_2;
    end
  end

endmodule
"""
  a._ref_src_yosys = a._ref_src
  do_test( a )

def test_if_dangling_else_outter( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_1 = InPort( Bits32 )
      s.in_2 = InPort( Bits32 )
      s.out = OutPort( Bits32 )
      @s.update
      def upblk():
        if Bits1(1):
          if Bits1(0):
            s.out = s.in_1
        else:
          s.out = s.in_2
  a = A()
  a._ref_src = \
"""
module A
(
  input logic [0:0] clk,
  input logic [31:0] in_1,
  input logic [31:0] in_2,
  output logic [31:0] out,
  input logic [0:0] reset
);

  // @s.update
  // def upblk():
  //   if Bits1(1):
  //     if Bits1(0):
  //       s.out = s.in_1
  //   else:
  //     s.out = s.in_2

  always_comb begin : upblk
    if ( 1'd1 ) begin
      if ( 1'd0 ) begin
        out = in_1;
      end
    end
    else
      out = in_2;
  end

endmodule
"""
  a._ref_src_yosys = a._ref_src
  do_test( a )

def test_if_branches( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_1 = InPort( Bits32 )
      s.in_2 = InPort( Bits32 )
      s.in_3 = InPort( Bits32 )
      s.out = OutPort( Bits32 )
      @s.update
      def upblk():
        if Bits1(1):
          s.out = s.in_1
        elif Bits1(0):
          s.out = s.in_2
        else:
          s.out = s.in_3
  a = A()
  a._ref_src = \
"""
module A
(
  input logic [0:0] clk,
  input logic [31:0] in_1,
  input logic [31:0] in_2,
  input logic [31:0] in_3,
  output logic [31:0] out,
  input logic [0:0] reset
);

  // @s.update
  // def upblk():
  //   if Bits1(1):
  //     s.out = s.in_1
  //   elif Bits1(0):
  //     s.out = s.in_2
  //   else:
  //     s.out = s.in_3

  always_comb begin : upblk
    if ( 1'd1 ) begin
      out = in_1;
    end
    else if ( 1'd0 ) begin
      out = in_2;
    end
    else
      out = in_3;
  end

endmodule
"""
  a._ref_src_yosys = a._ref_src
  do_test( a )

def test_nested_if( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_1 = InPort( Bits32 )
      s.in_2 = InPort( Bits32 )
      s.in_3 = InPort( Bits32 )
      s.out = OutPort( Bits32 )
      @s.update
      def upblk():
        if Bits1(1):
          if Bits1(0):
            s.out = s.in_1
          else:
            s.out = s.in_2
        elif Bits1(0):
          if Bits1(1):
            s.out = s.in_2
          else:
            s.out = s.in_3
        else:
          if Bits1(1):
            s.out = s.in_3
          else:
            s.out = s.in_1
  a = A()
  a._ref_src = \
"""
module A
(
  input logic [0:0] clk,
  input logic [31:0] in_1,
  input logic [31:0] in_2,
  input logic [31:0] in_3,
  output logic [31:0] out,
  input logic [0:0] reset
);

  // @s.update
  // def upblk():
  //   if Bits1(1):
  //     if Bits1(0):
  //       s.out = s.in_1
  //     else:
  //       s.out = s.in_2
  //   elif Bits1(0):
  //     if Bits1(1):
  //       s.out = s.in_2
  //     else:
  //       s.out = s.in_3
  //   else:
  //     if Bits1(1):
  //       s.out = s.in_3
  //     else:
  //       s.out = s.in_1

  always_comb begin : upblk
    if ( 1'd1 ) begin
      if ( 1'd0 ) begin
        out = in_1;
      end
      else
        out = in_2;
    end
    else if ( 1'd0 ) begin
      if ( 1'd1 ) begin
        out = in_2;
      end
      else
        out = in_3;
    end
    else if ( 1'd1 ) begin
      out = in_3;
    end
    else
      out = in_1;
  end

endmodule
"""
  a._ref_src_yosys = a._ref_src
  do_test( a )

def test_for_range_upper( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = [ InPort( Bits32 ) for _ in range(2) ]
      s.out = [ OutPort( Bits32 ) for _ in range(2) ]
      @s.update
      def upblk():
        for i in range(2):
          s.out[i] = s.in_[i]
  a = A()
  a._ref_src = \
"""
module A
(
  input logic [0:0] clk,
  input logic [31:0] in_ [0:1],
  output logic [31:0] out [0:1],
  input logic [0:0] reset
);

  // @s.update
  // def upblk():
  //   for i in range(2):
  //     s.out[i] = s.in_[i]

  always_comb begin : upblk
    for ( int i = 0; i < 2; i += 1 )
      out[i] = in_[i];
  end

endmodule
"""
  a._ref_src_yosys = \
"""
module A
(
  input logic [0:0] clk,
  input logic [31:0] in___0,
  input logic [31:0] in___1,
  output logic [31:0] out__0,
  output logic [31:0] out__1,
  input logic [0:0] reset
);
  logic [31:0] in_ [0:1];
  logic [31:0] out [0:1];

  // @s.update
  // def upblk():
  //   for i in range(2):
  //     s.out[i] = s.in_[i]

  integer __loopvar__upblk_i;

  always_comb begin : upblk
    for ( __loopvar__upblk_i = 0; __loopvar__upblk_i < 2; __loopvar__upblk_i = __loopvar__upblk_i + 1 )
      out[__loopvar__upblk_i] = in_[__loopvar__upblk_i];
  end

  assign in_[0] = in___0;
  assign in_[1] = in___1;
  assign out__0 = out[0];
  assign out__1 = out[1];

endmodule
"""
  do_test( a )

def test_for_range_lower_upper( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = [ InPort( Bits32 ) for _ in range(2) ]
      s.out = [ OutPort( Bits32 ) for _ in range(2) ]
      @s.update
      def upblk():
        for i in range(1, 2):
          s.out[i] = s.in_[i]
        s.out[0] = s.in_[0]
  a = A()
  a._ref_src = \
"""
module A
(
  input logic [0:0] clk,
  input logic [31:0] in_ [0:1],
  output logic [31:0] out [0:1],
  input logic [0:0] reset
);

  // @s.update
  // def upblk():
  //   for i in range(1, 2):
  //     s.out[i] = s.in_[i]
  //   s.out[0] = s.in_[0]

  always_comb begin : upblk
    for ( int i = 1; i < 2; i += 1 )
      out[i] = in_[i];
    out[0] = in_[0];
  end

endmodule
"""
  a._ref_src_yosys = \
"""
module A
(
  input logic [0:0] clk,
  input logic [31:0] in___0,
  input logic [31:0] in___1,
  output logic [31:0] out__0,
  output logic [31:0] out__1,
  input logic [0:0] reset
);
  logic [31:0] in_ [0:1];
  logic [31:0] out [0:1];

  // @s.update
  // def upblk():
  //   for i in range(1, 2):
  //     s.out[i] = s.in_[i]
  //   s.out[0] = s.in_[0]

  integer __loopvar__upblk_i;

  always_comb begin : upblk
    for ( __loopvar__upblk_i = 1; __loopvar__upblk_i < 2; __loopvar__upblk_i = __loopvar__upblk_i + 1 )
      out[__loopvar__upblk_i] = in_[__loopvar__upblk_i];
    out[0] = in_[0];
  end

  assign in_[0] = in___0;
  assign in_[1] = in___1;
  assign out__0 = out[0];
  assign out__1 = out[1];

endmodule
"""
  do_test( a )

def test_for_range_lower_upper_step( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = [ InPort( Bits32 ) for _ in range(5) ]
      s.out = [ OutPort( Bits32 ) for _ in range(5) ]
      @s.update
      def upblk():
        for i in range(0, 5, 2):
          s.out[i] = s.in_[i]
        for i in range(1, 5, 2):
          s.out[i] = s.in_[i]
  a = A()
  a._ref_src = \
"""
module A
(
  input logic [0:0] clk,
  input logic [31:0] in_ [0:4],
  output logic [31:0] out [0:4],
  input logic [0:0] reset
);

  // @s.update
  // def upblk():
  //   for i in range(0, 5, 2):
  //     s.out[i] = s.in_[i]
  //   for i in range(1, 5, 2):
  //     s.out[i] = s.in_[i]

  always_comb begin : upblk
    for ( int i = 0; i < 5; i += 2 )
      out[i] = in_[i];
    for ( int i = 1; i < 5; i += 2 )
      out[i] = in_[i];
  end

endmodule
"""
  a._ref_src_yosys = \
"""
module A
(
  input logic [0:0] clk,
  input logic [31:0] in___0,
  input logic [31:0] in___1,
  input logic [31:0] in___2,
  input logic [31:0] in___3,
  input logic [31:0] in___4,
  output logic [31:0] out__0,
  output logic [31:0] out__1,
  output logic [31:0] out__2,
  output logic [31:0] out__3,
  output logic [31:0] out__4,
  input logic [0:0] reset
);
  logic [31:0] in_ [0:4];
  logic [31:0] out [0:4];

  // @s.update
  // def upblk():
  //   for i in range(0, 5, 2):
  //     s.out[i] = s.in_[i]
  //   for i in range(1, 5, 2):
  //     s.out[i] = s.in_[i]

  integer __loopvar__upblk_i;

  always_comb begin : upblk
    for ( __loopvar__upblk_i = 0; __loopvar__upblk_i < 5; __loopvar__upblk_i = __loopvar__upblk_i + 2 )
      out[__loopvar__upblk_i] = in_[__loopvar__upblk_i];
    for ( __loopvar__upblk_i = 1; __loopvar__upblk_i < 5; __loopvar__upblk_i = __loopvar__upblk_i + 2 )
      out[__loopvar__upblk_i] = in_[__loopvar__upblk_i];
  end

  assign in_[0] = in___0;
  assign in_[1] = in___1;
  assign in_[2] = in___2;
  assign in_[3] = in___3;
  assign in_[4] = in___4;
  assign out__0 = out[0];
  assign out__1 = out[1];
  assign out__2 = out[2];
  assign out__3 = out[3];
  assign out__4 = out[4];

endmodule
"""
  do_test( a )

def test_if_exp_for( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = [ InPort( Bits32 ) for _ in range(5) ]
      s.out = [ OutPort( Bits32 ) for _ in range(5) ]
      @s.update
      def upblk():
        for i in range(5):
          s.out[i] = s.in_[i] if i == 1 else s.in_[0]
  a = A()
  a._ref_src = \
"""
module A
(
  input logic [0:0] clk,
  input logic [31:0] in_ [0:4],
  output logic [31:0] out [0:4],
  input logic [0:0] reset
);

  // @s.update
  // def upblk():
  //   for i in range(5):
  //     s.out[i] = s.in_[i] if i == 1 else s.in_[0]

  always_comb begin : upblk
    for ( int i = 0; i < 5; i += 1 )
      out[i] = ( i == 1 ) ? in_[i] : in_[0];
  end

endmodule
"""
  a._ref_src_yosys = \
"""
module A
(
  input logic [0:0] clk,
  input logic [31:0] in___0,
  input logic [31:0] in___1,
  input logic [31:0] in___2,
  input logic [31:0] in___3,
  input logic [31:0] in___4,
  output logic [31:0] out__0,
  output logic [31:0] out__1,
  output logic [31:0] out__2,
  output logic [31:0] out__3,
  output logic [31:0] out__4,
  input logic [0:0] reset
);
  logic [31:0] in_ [0:4];
  logic [31:0] out [0:4];

  // @s.update
  // def upblk():
  //   for i in range(5):
  //     s.out[i] = s.in_[i] if i == 1 else s.in_[0]

  integer __loopvar__upblk_i;

  always_comb begin : upblk
    for ( __loopvar__upblk_i = 0; __loopvar__upblk_i < 5; __loopvar__upblk_i = __loopvar__upblk_i + 1 )
      out[__loopvar__upblk_i] = ( __loopvar__upblk_i == 1 ) ? in_[__loopvar__upblk_i] : in_[0];
  end

  assign in_[0] = in___0;
  assign in_[1] = in___1;
  assign in_[2] = in___2;
  assign in_[3] = in___3;
  assign in_[4] = in___4;
  assign out__0 = out[0];
  assign out__1 = out[1];
  assign out__2 = out[2];
  assign out__3 = out[3];
  assign out__4 = out[4];

endmodule
"""
  do_test( a )

def test_if_exp_unary_op( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = [ InPort( Bits32 ) for _ in range(5) ]
      s.out = [ OutPort( Bits32 ) for _ in range(5) ]
      @s.update
      def upblk():
        for i in range(5):
          s.out[i] = (~s.in_[i]) if i == 1 else s.in_[0]
  a = A()
  a._ref_src = \
"""
module A
(
  input logic [0:0] clk,
  input logic [31:0] in_ [0:4],
  output logic [31:0] out [0:4],
  input logic [0:0] reset
);

  // @s.update
  // def upblk():
  //   for i in range(5):
  //     s.out[i] = (~s.in_[i]) if i == 1 else s.in_[0]

  always_comb begin : upblk
    for ( int i = 0; i < 5; i += 1 )
      out[i] = ( i == 1 ) ? ~in_[i] : in_[0];
  end

endmodule
"""
  a._ref_src_yosys = \
"""
module A
(
  input logic [0:0] clk,
  input logic [31:0] in___0,
  input logic [31:0] in___1,
  input logic [31:0] in___2,
  input logic [31:0] in___3,
  input logic [31:0] in___4,
  output logic [31:0] out__0,
  output logic [31:0] out__1,
  output logic [31:0] out__2,
  output logic [31:0] out__3,
  output logic [31:0] out__4,
  input logic [0:0] reset
);
  logic [31:0] in_ [0:4];
  logic [31:0] out [0:4];

  // @s.update
  // def upblk():
  //   for i in range(5):
  //     s.out[i] = (~s.in_[i]) if i == 1 else s.in_[0]

  integer __loopvar__upblk_i;

  always_comb begin : upblk
    for ( __loopvar__upblk_i = 0; __loopvar__upblk_i < 5; __loopvar__upblk_i = __loopvar__upblk_i + 1 )
      out[__loopvar__upblk_i] = ( __loopvar__upblk_i == 1 ) ? ~in_[__loopvar__upblk_i] : in_[0];
  end

  assign in_[0] = in___0;
  assign in_[1] = in___1;
  assign in_[2] = in___2;
  assign in_[3] = in___3;
  assign in_[4] = in___4;
  assign out__0 = out[0];
  assign out__1 = out[1];
  assign out__2 = out[2];
  assign out__3 = out[3];
  assign out__4 = out[4];

endmodule
"""
  do_test( a )

def test_if_bool_op( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = [ InPort( Bits32 ) for _ in range(5) ]
      s.out = [ OutPort( Bits32 ) for _ in range(5) ]
      @s.update
      def upblk():
        for i in range(5):
          if s.in_[i] and (s.in_[i+1] if i<5 else s.in_[4]):
            s.out[i] = s.in_[i]
          else:
            s.out[i] = Bits32(0)
  a = A()
  a._ref_src = \
"""
module A
(
  input logic [0:0] clk,
  input logic [31:0] in_ [0:4],
  output logic [31:0] out [0:4],
  input logic [0:0] reset
);

  // @s.update
  // def upblk():
  //   for i in range(5):
  //     if s.in_[i] and (s.in_[i+1] if i<5 else s.in_[4]):
  //       s.out[i] = s.in_[i]
  //     else:
  //       s.out[i] = Bits32(0)

  always_comb begin : upblk
    for ( int i = 0; i < 5; i += 1 )
      if ( in_[i] && ( ( i < 5 ) ? in_[i + 1] : in_[4] ) ) begin
        out[i] = in_[i];
      end
      else
        out[i] = 32'd0;
  end

endmodule
"""
  a._ref_src_yosys = \
"""
module A
(
  input logic [0:0] clk,
  input logic [31:0] in___0,
  input logic [31:0] in___1,
  input logic [31:0] in___2,
  input logic [31:0] in___3,
  input logic [31:0] in___4,
  output logic [31:0] out__0,
  output logic [31:0] out__1,
  output logic [31:0] out__2,
  output logic [31:0] out__3,
  output logic [31:0] out__4,
  input logic [0:0] reset
);
  logic [31:0] in_ [0:4];
  logic [31:0] out [0:4];

  // @s.update
  // def upblk():
  //   for i in range(5):
  //     if s.in_[i] and (s.in_[i+1] if i<5 else s.in_[4]):
  //       s.out[i] = s.in_[i]
  //     else:
  //       s.out[i] = Bits32(0)

  integer __loopvar__upblk_i;

  always_comb begin : upblk
    for ( __loopvar__upblk_i = 0; __loopvar__upblk_i < 5; __loopvar__upblk_i = __loopvar__upblk_i + 1 )
      if ( in_[__loopvar__upblk_i] && ( ( __loopvar__upblk_i < 5 ) ? in_[__loopvar__upblk_i + 1] : in_[4] ) ) begin
        out[__loopvar__upblk_i] = in_[__loopvar__upblk_i];
      end
      else
        out[__loopvar__upblk_i] = 32'd0;
  end

  assign in_[0] = in___0;
  assign in_[1] = in___1;
  assign in_[2] = in___2;
  assign in_[3] = in___3;
  assign in_[4] = in___4;
  assign out__0 = out[0];
  assign out__1 = out[1];
  assign out__2 = out[2];
  assign out__3 = out[3];
  assign out__4 = out[4];

endmodule
"""
  do_test( a )

def test_tmpvar( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = [ InPort( Bits32 ) for _ in range(5) ]
      s.out = [ OutPort( Bits32 ) for _ in range(5) ]
      @s.update
      def upblk():
        for i in range(5):
          if s.in_[i] and (s.in_[i+1] if i<5 else s.in_[4]):
            tmpvar = s.in_[i]
          else:
            tmpvar = Bits32(0)
          s.out[i] = tmpvar
  a = A()
  a._ref_src = \
"""
module A
(
  input logic [0:0] clk,
  input logic [31:0] in_ [0:4],
  output logic [31:0] out [0:4],
  input logic [0:0] reset
);
  logic [31:0] __tmpvar__upblk_tmpvar;

  // @s.update
  // def upblk():
  //   for i in range(5):
  //     if s.in_[i] and (s.in_[i+1] if i<5 else s.in_[4]):
  //       tmpvar = s.in_[i]
  //     else:
  //       tmpvar = Bits32(0)
  //     s.out[i] = tmpvar

  always_comb begin : upblk
    for ( int i = 0; i < 5; i += 1 ) begin
      if ( in_[i] && ( ( i < 5 ) ? in_[i + 1] : in_[4] ) ) begin
        __tmpvar__upblk_tmpvar = in_[i];
      end
      else
        __tmpvar__upblk_tmpvar = 32'd0;
      out[i] = __tmpvar__upblk_tmpvar;
    end
  end

endmodule
"""
  a._ref_src_yosys = \
"""
module A
(
  input logic [0:0] clk,
  input logic [31:0] in___0,
  input logic [31:0] in___1,
  input logic [31:0] in___2,
  input logic [31:0] in___3,
  input logic [31:0] in___4,
  output logic [31:0] out__0,
  output logic [31:0] out__1,
  output logic [31:0] out__2,
  output logic [31:0] out__3,
  output logic [31:0] out__4,
  input logic [0:0] reset
);
  logic [31:0] in_ [0:4];
  logic [31:0] out [0:4];
  logic [31:0] __tmpvar__upblk_tmpvar;

  // @s.update
  // def upblk():
  //   for i in range(5):
  //     if s.in_[i] and (s.in_[i+1] if i<5 else s.in_[4]):
  //       tmpvar = s.in_[i]
  //     else:
  //       tmpvar = Bits32(0)
  //     s.out[i] = tmpvar

  integer __loopvar__upblk_i;

  always_comb begin : upblk
    for ( __loopvar__upblk_i = 0; __loopvar__upblk_i < 5; __loopvar__upblk_i = __loopvar__upblk_i + 1 ) begin
      if ( in_[__loopvar__upblk_i] && ( ( __loopvar__upblk_i < 5 ) ? in_[__loopvar__upblk_i + 1] : in_[4] ) ) begin
        __tmpvar__upblk_tmpvar = in_[__loopvar__upblk_i];
      end
      else
        __tmpvar__upblk_tmpvar = 32'd0;
      out[__loopvar__upblk_i] = __tmpvar__upblk_tmpvar;
    end
  end

  assign in_[0] = in___0;
  assign in_[1] = in___1;
  assign in_[2] = in___2;
  assign in_[3] = in___3;
  assign in_[4] = in___4;
  assign out__0 = out[0];
  assign out__1 = out[1];
  assign out__2 = out[2];
  assign out__3 = out[3];
  assign out__4 = out[4];

endmodule
"""
  do_test( a )

def test_struct( do_test ):
  @bitstruct
  class B:
    foo: Bits32
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( B )
      s.out = OutPort( Bits32 )
      @s.update
      def upblk():
        s.out = s.in_.foo
  a = A()
  a._ref_src = \
"""
typedef struct packed {
  logic [31:0] foo;
} B;

module A
(
  input logic [0:0] clk,
  input B in_,
  output logic [31:0] out,
  input logic [0:0] reset
);

  // @s.update
  // def upblk():
  //   s.out = s.in_.foo

  always_comb begin : upblk
    out = in_.foo;
  end

endmodule
"""
  a._ref_src_yosys = \
"""
module A
(
  input logic [0:0] clk,
  input logic [31:0] in___foo,
  output logic [31:0] out,
  input logic [0:0] reset
);
  logic [31:0] in_;

  // @s.update
  // def upblk():
  //   s.out = s.in_.foo

  always_comb begin : upblk
    out = in___foo;
  end

  assign in_[31:0] = in___foo;

endmodule
"""
  do_test( a )

def test_packed_array_concat( do_test ):
  @bitstruct
  class B:
    bar: [ Bits32 ] * 2
    foo: Bits32
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( B )
      s.out = OutPort( Bits96 )
      @s.update
      def upblk():
        s.out = concat( s.in_.bar[0], s.in_.bar[1], s.in_.foo )
  a = A()
  a._ref_src = \
"""
typedef struct packed {
  logic [1:0][31:0] bar;
  logic [31:0] foo;
} B;

module A
(
  input logic [0:0] clk,
  input B in_,
  output logic [95:0] out,
  input logic [0:0] reset
);

  // @s.update
  // def upblk():
  //   s.out = concat( s.in_.bar[0], s.in_.bar[1], s.in_.foo )

  always_comb begin : upblk
    out = { in_.bar[0], in_.bar[1], in_.foo };
  end

endmodule
"""
  a._ref_src_yosys = \
"""
module A
(
  input logic [0:0] clk,
  input logic [31:0] in___bar__0,
  input logic [31:0] in___bar__1,
  input logic [31:0] in___foo,
  output logic [95:0] out,
  input logic [0:0] reset
);
  logic [31:0] in___bar [0:1];
  logic [95:0] in_;

  // @s.update
  // def upblk():
  //   s.out = concat( s.in_.bar[0], s.in_.bar[1], s.in_.foo )

  always_comb begin : upblk
    out = { in___bar[0], in___bar[1], in___foo };
  end

  assign in___bar[0] = in___bar__0;
  assign in___bar[1] = in___bar__1;
  assign in_[95:64] = in___bar__1;
  assign in_[63:32] = in___bar__0;
  assign in_[31:0] = in___foo;

endmodule
"""
  do_test( a )

def test_nested_struct( do_test ):
  @bitstruct
  class C:
    woof: Bits32
  @bitstruct
  class B:
    bar: [ Bits32 ]*2
    c: C
    foo: Bits32
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( B )
      s.out = OutPort( Bits96 )
      @s.update
      def upblk():
        s.out = concat( s.in_.bar[0], s.in_.c.woof, s.in_.foo )
  a = A()
  a._ref_src = \
"""
typedef struct packed {
  logic [31:0] woof;
} C;

typedef struct packed {
  logic [1:0][31:0] bar;
  C c;
  logic [31:0] foo;
} B;

module A
(
  input logic [0:0] clk,
  input B in_,
  output logic [95:0] out,
  input logic [0:0] reset
);

  // @s.update
  // def upblk():
  //   s.out = concat( s.in_.bar[0], s.in_.c.woof, s.in_.foo )

  always_comb begin : upblk
    out = { in_.bar[0], in_.c.woof, in_.foo };
  end

endmodule
"""
  a._ref_src_yosys = \
"""
module A
(
  input logic [0:0] clk,
  input logic [31:0] in___bar__0,
  input logic [31:0] in___bar__1,
  input logic [31:0] in___c__woof,
  input logic [31:0] in___foo,
  output logic [95:0] out,
  input logic [0:0] reset
);
  logic [31:0] in___bar [0:1];
  logic [31:0] in___c;
  logic [127:0] in_;

  // @s.update
  // def upblk():
  //   s.out = concat( s.in_.bar[0], s.in_.c.woof, s.in_.foo )

  always_comb begin : upblk
    out = { in___bar[0], in___c__woof, in___foo };
  end

  assign in___bar[0] = in___bar__0;
  assign in___bar[1] = in___bar__1;
  assign in___c[31:0] = in___c__woof;
  assign in_[127:96] = in___bar__1;
  assign in_[95:64] = in___bar__0;
  assign in_[63:32] = in___c__woof;
  assign in_[31:0] = in___foo;

endmodule
"""
  do_test( a )

def test_lambda_connect( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.out = OutPort( Bits32 )
      s.out //= lambda: s.in_ + Bits32(42)
  a = A()
  a._ref_src = \
"""
module A
(
  input logic [0:0] clk,
  input logic [31:0] in_,
  output logic [31:0] out,
  input logic [0:0] reset
);

  // def _lambda__s_out(): s.out = s.in_ + Bits32(42)

  always_comb begin : _lambda__s_out
    out = in_ + 32'd42;
  end

endmodule
"""
  a._ref_src_yosys = a._ref_src
  do_test( a )

#-------------------------------------------------------------------------
# Structural
#-------------------------------------------------------------------------

def test_struct_port( do_test ):
  @bitstruct
  class B:
    foo: Bits32
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( B )
      s.out = OutPort( Bits32 )
      connect( s.out, s.in_.foo )
  a = A()
  a._ref_src = \
"""
typedef struct packed {
  logic [31:0] foo;
} B;

module A
(
  input logic [0:0] clk,
  input B in_,
  output logic [31:0] out,
  input logic [0:0] reset
);

  assign out = in_.foo;

endmodule
"""
  a._ref_src_yosys = \
"""
module A
(
  input logic [0:0] clk,
  input logic [31:0] in___foo,
  output logic [31:0] out,
  input logic [0:0] reset
);
  logic [31:0] in_;

  assign in_[31:0] = in___foo;
  assign out = in___foo;

endmodule
"""
  do_test( a )

def test_nested_struct_port( do_test ):
  @bitstruct
  class C:
    bar: Bits32
  @bitstruct
  class B:
    foo: Bits32
    c: C
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( B )
      s.out_foo = OutPort( Bits32 )
      s.out_bar = OutPort( Bits32 )
      connect( s.out_foo, s.in_.foo )
      connect( s.out_bar, s.in_.c.bar )
  a = A()
  a._ref_src = \
"""
typedef struct packed {
  logic [31:0] bar;
} C;

typedef struct packed {
  logic [31:0] foo;
  C c;
} B;

module A
(
  input logic [0:0] clk,
  input B in_,
  output logic [31:0] out_bar,
  output logic [31:0] out_foo,
  input logic [0:0] reset
);

  assign out_foo = in_.foo;
  assign out_bar = in_.c.bar;

endmodule
"""
  a._ref_src_yosys = \
"""
module A
(
  input logic [0:0] clk,
  input logic [31:0] in___foo,
  input logic [31:0] in___c__bar,
  output logic [31:0] out_bar,
  output logic [31:0] out_foo,
  input logic [0:0] reset
);
  logic [31:0] in___c;
  logic [63:0] in_;

  assign in___c[31:0] = in___c__bar;
  assign in_[63:32] = in___foo;
  assign in_[31:0] = in___c__bar;
  assign out_foo = in___foo;
  assign out_bar = in___c__bar;

endmodule
"""
  do_test( a )

def test_packed_array( do_test ):
  @bitstruct
  class B:
    foo: [ Bits32 ] * 2
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( B )
      s.out =  [ OutPort( Bits32 ) for _ in range(2) ]
      connect( s.out[0], s.in_.foo[0] )
      connect( s.out[1], s.in_.foo[1] )
  a = A()
  a._ref_src = \
"""
typedef struct packed {
  logic [1:0][31:0] foo;
} B;

module A
(
  input logic [0:0] clk,
  input B in_,
  output logic [31:0] out [0:1],
  input logic [0:0] reset
);

  assign out[0] = in_.foo[0];
  assign out[1] = in_.foo[1];

endmodule
"""
  a._ref_src_yosys = \
"""
module A
(
  input logic [0:0] clk,
  input logic [31:0] in___foo__0,
  input logic [31:0] in___foo__1,
  output logic [31:0] out__0,
  output logic [31:0] out__1,
  input logic [0:0] reset
);
  logic [31:0] in___foo [0:1];
  logic [63:0] in_;
  logic [31:0] out [0:1];

  assign in___foo[0] = in___foo__0;
  assign in___foo[1] = in___foo__1;
  assign in_[63:32] = in___foo__1;
  assign in_[31:0] = in___foo__0;
  assign out__0 = out[0];
  assign out__1 = out[1];
  assign out[0] = in___foo[0];
  assign out[1] = in___foo[1];

endmodule
"""
  do_test( a )

def test_struct_packed_array( do_test ):
  @bitstruct
  class C:
    bar: Bits32
  @bitstruct
  class B:
    c: [ C ] * 2
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( B )
      s.out =  [ OutPort( Bits32 ) for _ in range(2) ]
      connect( s.out[0], s.in_.c[0].bar )
      connect( s.out[1], s.in_.c[1].bar )
  a = A()
  a._ref_src = \
"""
typedef struct packed {
  logic [31:0] bar;
} C;

typedef struct packed {
  C [1:0] c;
} B;

module A
(
  input logic [0:0] clk,
  input B in_,
  output logic [31:0] out [0:1],
  input logic [0:0] reset
);

  assign out[0] = in_.c[0].bar;
  assign out[1] = in_.c[1].bar;

endmodule
"""
  a._ref_src_yosys = \
"""
module A
(
  input logic [0:0] clk,
  input logic [31:0] in___c__0__bar,
  input logic [31:0] in___c__1__bar,
  output logic [31:0] out__0,
  output logic [31:0] out__1,
  input logic [0:0] reset
);
  logic [31:0] in___c__bar [0:1];
  logic [31:0] in___c [0:1];
  logic [63:0] in_;
  logic [31:0] out [0:1];

  assign in___c__bar[0] = in___c__0__bar;
  assign in___c[0][31:0] = in___c__0__bar;
  assign in___c__bar[1] = in___c__1__bar;
  assign in___c[1][31:0] = in___c__1__bar;
  assign in_[63:32] = in___c__1__bar;
  assign in_[31:0] = in___c__0__bar;
  assign out__0 = out[0];
  assign out__1 = out[1];
  assign out[0] = in___c__bar[0];
  assign out[1] = in___c__bar[1];

endmodule
"""
  do_test( a )

def test_long_component_name( do_test ):
  @bitstruct
  class ThisIsABitStructWithSuperLongName:
    bar: Bits32
  class A( Component ):
    def construct( s, T1, T2, T3, T4, T5, T6, T7 ):
      s.in_ = InPort( Bits32 )
      s.wire_ = Wire( Bits32 )
      s.out = OutPort( Bits32 )
      connect( s.in_, s.wire_ )
      connect( s.wire_, s.out )
  args = [ThisIsABitStructWithSuperLongName]*7
  a = A(*args)
  a._ref_src = \
"""
module A__a840bd1c84c05ea2
(
  input logic [0:0] clk,
  input logic [31:0] in_,
  output logic [31:0] out,
  input logic [0:0] reset
);
  logic [31:0] wire_;

  assign wire_ = in_;
  assign out = wire_;

endmodule
"""
  a._ref_src_yosys = a._ref_src
  do_test( a )

def test_const_struct( do_test ):
  @bitstruct
  class ST:
    foo: Bits16
    bar: Bits32

  @bitstruct
  class COMB:
    fst: ST
    snd: ST

  class Top( Component ):
    def construct( s ):
      s.out = OutPort( COMB )
      connect( s.out, COMB(ST(1, 2), ST(3, 4)) )
  a = Top()
  a._ref_src = \
"""
typedef struct packed {
  logic [15:0] foo;
  logic [31:0] bar;
} ST;

typedef struct packed {
  ST fst;
  ST snd;
} COMB;

module Top
(
  input logic [0:0] clk,
  output COMB out,
  input logic [0:0] reset
);

  assign out = { { 16'd1, 32'd2 }, { 16'd3, 32'd4 } };

endmodule
"""
  a._ref_src_yosys = \
"""
module Top
(
  input logic [0:0] clk,
  output logic [15:0] out__fst__foo,
  output logic [31:0] out__fst__bar,
  output logic [15:0] out__snd__foo,
  output logic [31:0] out__snd__bar,
  input logic [0:0] reset
);
  logic [47:0]  out__fst;
  logic [47:0]  out__snd;
  logic [95:0]  out;

  assign out__fst__foo = out__fst[47:32];
  assign out__fst__bar = out__fst[31:0];
  assign out__snd__foo = out__snd[47:32];
  assign out__snd__bar = out__snd[31:0];
  assign out__fst__foo = out[95:80];
  assign out__fst__bar = out[79:48];
  assign out__snd__foo = out[47:32];
  assign out__snd__bar = out[31:0];
  assign out = { { 16'd1, 32'd2 }, { 16'd3, 32'd4 } };

endmodule
"""
  do_test( a )

@pytest.mark.xfail(run=False, reason="TODO: resolving BitStructs according to name AND fields")
def test_struct_uniqueness( do_test ):
  class A:
    @bitstruct
    class ST:
      a_foo: Bits16
      a_bar: Bits32

  class B:
    @bitstruct
    class ST:
      b_foo: Bits16
      b_bar: Bits32

  @bitstruct
  class COMB:
    fst: A.ST
    snd: B.ST

  class Top( Component ):
    def construct( s ):
      s.out = OutPort( COMB )
      connect( s.out, COMB(A.ST(1, 2), B.ST(3, 4)) )
  a = Top()
  a._ref_src = \
"""
typedef struct packed {
  logic [15:0] foo;
  logic [31:0] bar;
} ST;

typedef struct packed {
  ST fst;
  ST snd;
} COMB;

module Top
(
  input logic [0:0] clk,
  output COMB out,
  input logic [0:0] reset
);

  assign out = { { 16'd1, 32'd2 }, { 16'd3, 32'd4 } };

endmodule
"""
  a._ref_src_yosys = \
"""
module Top
(
  input logic [0:0] clk,
  output logic [15:0] out__fst__foo,
  output logic [31:0] out__fst__bar,
  output logic [15:0] out__snd__foo,
  output logic [31:0] out__snd__bar,
  input logic [0:0] reset
);
  logic [47:0]  out__fst;
  logic [47:0]  out__snd;
  logic [95:0]  out;

  assign out__fst__foo = out__fst[47:32];
  assign out__fst__bar = out__fst[31:0];
  assign out__snd__foo = out__snd[47:32];
  assign out__snd__bar = out__snd[31:0];
  assign out__fst__foo = out[95:80];
  assign out__fst__bar = out[79:48];
  assign out__snd__foo = out[47:32];
  assign out__snd__bar = out[31:0];
  assign out = { { 16'd1, 32'd2 }, { 16'd3, 32'd4 } };

endmodule
"""
  do_test( a )



#=========================================================================
# SVTranslator_L4_cases_test.py
#=========================================================================
"""Test the SystemVerilog translator."""

from pymtl3.datatypes import Bits32, Bits64, concat, sext, zext
from pymtl3.dsl import Component, InPort, OutPort, connect
from pymtl3.passes.rtlir.util.test_utility import do_test

from ..structural.test.SVStructuralTranslatorL1_test import check_eq
from ..SVTranslator import SVTranslator


def local_do_test( m ):
  m.elaborate()
  tr = SVTranslator( m )
  tr.translate( m )
  check_eq( tr.hierarchy.src, m._ref_src )

def test_subcomponent( do_test ):
  class B( Component ):
    def construct( s ):
      s.foo = OutPort( Bits32 )
      @s.update
      def upblk():
        s.foo = Bits32(42)
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits32 )
      s.b = B()
      @s.update
      def upblk():
        s.out = s.b.foo
  a = A()
  a._ref_src = \
"""
module B
(
  input logic [0:0] clk,
  output logic [31:0] foo,
  input logic [0:0] reset
);

  // @s.update
  // def upblk():
  //   s.foo = Bits32(42)

  always_comb begin : upblk
    foo = 32'd42;
  end

endmodule


module A
(
  input logic [0:0] clk,
  output logic [31:0] out,
  input logic [0:0] reset
);
  logic [0:0] b__clk;
  logic [31:0] b__foo;
  logic [0:0] b__reset;

  B b
  (
    .clk( b__clk ),
    .foo( b__foo ),
    .reset( b__reset )
  );

  // @s.update
  // def upblk():
  //   s.out = s.b.foo

  always_comb begin : upblk
    out = b__foo;
  end

  assign b__clk = clk;
  assign b__reset = reset;

endmodule
"""
  a._ref_src_yosys = a._ref_src
  do_test( a )

def test_sub_component_attr( do_test ):
  class B( Component ):
    def construct( s ):
      s.out_b = OutPort( Bits32 )
      @s.update
      def upblk():
        s.out_b = Bits32(0)
  class A( Component ):
    def construct( s ):
      s.b = B()
      s.out = OutPort( Bits64 )
      @s.update
      def upblk():
        s.out = zext( s.b.out_b, 64 )
  a = A()
  a._ref_src = \
"""
module B
(
  input logic [0:0] clk,
  output logic [31:0] out_b,
  input logic [0:0] reset
);

  // @s.update
  // def upblk():
  //   s.out_b = Bits32(0)

  always_comb begin : upblk
    out_b = 32'd0;
  end

endmodule


module A
(
  input logic [0:0] clk,
  output logic [63:0] out,
  input logic [0:0] reset
);
  logic [0:0] b__clk;
  logic [31:0] b__out_b;
  logic [0:0] b__reset;

  B b
  (
    .clk( b__clk ),
    .out_b( b__out_b ),
    .reset( b__reset )
  );

  // @s.update
  // def upblk():
  //   s.out = zext( s.b.out_b, 64 )

  always_comb begin : upblk
    out = { { 32 { 1'b0 } }, b__out_b };
  end

  assign b__clk = clk;
  assign b__reset = reset;

endmodule
"""
  a._ref_src_yosys = a._ref_src
  do_test( a )

def test_subcomponent_index( do_test ):
  class B( Component ):
    def construct( s ):
      s.out = OutPort( Bits32 )
  class A( Component ):
    def construct( s ):
      s.comp = [ B() for _ in range(2) ]
      s.out = OutPort( Bits32 )
      @s.update
      def upblk():
        s.out = s.comp[1].out
  a = A()
  a._ref_src = \
"""
module B
(
  input logic [0:0] clk,
  output logic [31:0] out,
  input logic [0:0] reset
);

endmodule


module A
(
  input logic [0:0] clk,
  output logic [31:0] out,
  input logic [0:0] reset
);
  logic [0:0] comp__0__clk;
  logic [31:0] comp__0__out;
  logic [0:0] comp__0__reset;

  B comp__0
  (
    .clk( comp__0__clk ),
    .out( comp__0__out ),
    .reset( comp__0__reset )
  );

  logic [0:0] comp__1__clk;
  logic [31:0] comp__1__out;
  logic [0:0] comp__1__reset;

  B comp__1
  (
    .clk( comp__1__clk ),
    .out( comp__1__out ),
    .reset( comp__1__reset )
  );

  // @s.update
  // def upblk():
  //   s.out = s.comp[1].out

  always_comb begin : upblk
    out = comp__1__out;
  end

  assign comp__0__clk = clk;
  assign comp__0__reset = reset;
  assign comp__1__clk = clk;
  assign comp__1__reset = reset;

endmodule
"""
  a._ref_src_yosys = \
"""
module B
(
  input logic [0:0] clk,
  output logic [31:0] out,
  input logic [0:0] reset
);

endmodule


module A
(
  input logic [0:0] clk,
  output logic [31:0] out,
  input logic [0:0] reset
);
  logic [0:0] comp__clk [0:1];
  logic [31:0] comp__out [0:1];
  logic [0:0] comp__reset [0:1];
  logic [0:0] comp__0__clk;
  logic [31:0] comp__0__out;
  logic [0:0] comp__0__reset;

  B comp__0
  (
    .clk( comp__0__clk ),
    .out( comp__0__out ),
    .reset( comp__0__reset )
  );

  logic [0:0] comp__1__clk;
  logic [31:0] comp__1__out;
  logic [0:0] comp__1__reset;

  B comp__1
  (
    .clk( comp__1__clk ),
    .out( comp__1__out ),
    .reset( comp__1__reset )
  );
  assign comp__0__clk = comp__clk[0];
  assign comp__1__clk = comp__clk[1];
  assign comp__out[0] = comp__0__out;
  assign comp__out[1] = comp__1__out;
  assign comp__0__reset = comp__reset[0];
  assign comp__1__reset = comp__reset[1];

  // @s.update
  // def upblk():
  //   s.out = s.comp[1].out

  always_comb begin : upblk
    out = comp__out[1];
  end

  assign comp__clk[0] = clk;
  assign comp__reset[0] = reset;
  assign comp__clk[1] = clk;
  assign comp__reset[1] = reset;

endmodule
"""
  do_test( a )



#=========================================================================
# BehavioralRTLIRPass_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : Feb 2, 2019
"""Test the behavioral RTLIR passes.

The generation, type check, and visualization passes are invoked. The
results of generation pass are verifed against a reference AST.
"""

import pytest

from pymtl3.datatypes import *
from pymtl3.dsl import Component, InPort, OutPort, connect
from pymtl3.passes.rtlir.behavioral import (
    BehavioralRTLIRGenPass,
    BehavioralRTLIRTypeCheckPass,
    BehavioralRTLIRVisualizationPass,
)
from pymtl3.passes.rtlir.behavioral.BehavioralRTLIR import *
from pymtl3.passes.rtlir.behavioral.test.BehavioralRTLIRL1Pass_test import XFAIL_ON_PY3
from pymtl3.passes.rtlir.errors import PyMTLTypeError
from pymtl3.passes.rtlir.util.test_utility import do_test, expected_failure


def local_do_test( m ):
  """Check if generated behavioral RTLIR is the same as reference."""

  ref = m._rtlir_test_ref
  m.elaborate()
  m.apply( BehavioralRTLIRGenPass() )
  m.apply( BehavioralRTLIRTypeCheckPass() )
  m.apply( BehavioralRTLIRVisualizationPass() )

  for blk in m.get_update_blocks():
    assert\
      m._pass_behavioral_rtlir_gen.rtlir_upblks[ blk ] == ref[ blk.__name__ ]

def test_reduce( do_test ):
  class v_reduce( Component ):
    def construct( s ):
      s.in_1 = InPort( Bits32 )
      s.in_2 = InPort( Bits32 )
      s.in_3 = InPort( Bits32 )
      s.out = OutPort( Bits1 )
      @s.update
      def v_reduce():
        s.out = reduce_and( s.in_1 ) & reduce_or( s.in_2 ) | reduce_xor( s.in_3 )

  a = v_reduce()
  in_1 = Attribute( Base( a ), 'in_1' )
  in_2 = Attribute( Base( a ), 'in_2' )
  in_3 = Attribute( Base( a ), 'in_3' )
  out = Attribute( Base( a ), 'out' )

  a._rtlir_test_ref = { 'v_reduce' : CombUpblk( 'v_reduce', [
    Assign( out, BinOp(
      BinOp( Reduce( BitAnd(), in_1 ), BitAnd(), Reduce( BitOr(), in_2 ) ),
      BitOr(), Reduce( BitXor(), in_3 ),
    ), True )
  ] ) }

  a._test_vector = [
          ' in_1        in_2     in_3     *out',
    [     Bits32,    Bits32,   Bits32,   Bits1    ],

    [          0,         1,        2,        1   ],
    [    b32(-1),         1,  b32(-1),        1   ],
    [          9,         8,        7,        1   ],
  ]

  do_test( a )

def test_index_basic( do_test ):
  class index_basic( Component ):
    def construct( s ):
      s.in_ = [ InPort( Bits16 ) for _ in range( 4 ) ]
      s.out = [ OutPort( Bits16 ) for _ in range( 2 ) ]

      @s.update
      def index_basic():
        s.out[ 0 ] = s.in_[ 0 ] + s.in_[ 1 ]
        s.out[ 1 ] = s.in_[ 2 ] + s.in_[ 3 ]

  a = index_basic()

  a._rtlir_test_ref = { 'index_basic' : CombUpblk( 'index_basic', [
    Assign( Index( Attribute( Base( a ), 'out' ), Number( 0 ) ),
      BinOp( Index( Attribute( Base( a ), 'in_' ), Number( 0 ) ), Add(),
             Index( Attribute( Base( a ), 'in_' ), Number( 1 ) ) ), True ),
    Assign( Index( Attribute( Base( a ), 'out' ), Number( 1 ) ),
      BinOp( Index( Attribute( Base( a ), 'in_' ), Number( 2 ) ), Add(),
             Index( Attribute( Base( a ), 'in_' ), Number( 3 ) ) ), True )
  ] ) }

  a._test_vector = [
          'in_[0]     in_[1]    in_[2]    in_[3]    *out[0]     *out[1]',
    [     Bits16,    Bits16,   Bits16,   Bits16,    Bits16,     Bits16 ],

    [          0,         1,        2,        3,         1,          5 ],
    [ Bits16(-1),         1, Bits16(-1),      1,         0,          0 ],
    [          9,         8,        7,        6,        17,         13 ],
  ]

  do_test( a )

def test_mismatch_width_assign( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( Bits16 )
      s.out = OutPort( Bits8 )

      @s.update
      def mismatch_width_assign():
        s.out = s.in_

  a = A()

  a._rtlir_test_ref = { 'mismatch_width_assign' : CombUpblk(
    'mismatch_width_assign', [ Assign(
      Attribute( Base( a ), 'out' ), Attribute( Base( a ), 'in_' ), True
    )
  ] ) }

  a._test_vector = [
                'in_             *out',
    [        Bits16,           Bits8 ],

    [             0,               0 ],
    [             2,               2 ],
    [    Bits16(-1),       Bits8(-1) ],
    [    Bits16(-2),       Bits8(-2) ],
    [ Bits16(32767),      Bits8(255) ],
  ]

  with expected_failure( PyMTLTypeError, "LHS and RHS of assignment" ):
    do_test( a )

def test_slicing_basic( do_test ):
  class slicing_basic( Component ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.out = OutPort( Bits64 )

      @s.update
      def slicing_basic():
        s.out[ 0:16 ] = s.in_[ 16:32 ]
        s.out[ 16:32 ] = s.in_[ 0:16 ]

  a = slicing_basic()

  a._rtlir_test_ref = { 'slicing_basic' : CombUpblk( 'slicing_basic', [
    Assign( Slice( Attribute( Base( a ), 'out' ), Number( 0 ), Number( 16 ) ),
      Slice( Attribute( Base( a ), 'in_' ), Number( 16 ), Number( 32 ) ), True ),
    Assign( Slice( Attribute( Base( a ), 'out' ), Number( 16 ), Number( 32 ) ),
      Slice( Attribute( Base( a ), 'in_' ), Number( 0 ), Number( 16 ) ), True )
  ] ) }

  a._test_vector = [
                'in_                        *out',
    [        Bits32,                     Bits64 ],

    [             0,                          0 ],
    [             2,            Bits64(0x20000) ],
    [    Bits32(-1),         Bits64(0xffffffff) ],
    [    Bits32(-2),         Bits64(0xfffeffff) ],
    [ Bits32(32767),         Bits64(0x7fff0000) ],
  ]

  do_test( a )

def test_bits_basic( do_test ):
  class bits_basic( Component ):
    def construct( s ):
      s.in_ = InPort( Bits16 )
      s.out = OutPort( Bits16 )

      @s.update
      def bits_basic():
        s.out = s.in_ + Bits16( 10 )

  a = bits_basic()

  a._rtlir_test_ref = { 'bits_basic' : CombUpblk( 'bits_basic', [
    Assign( Attribute( Base( a ), 'out' ),
      BinOp( Attribute( Base( a ), 'in_' ), Add(), SizeCast( 16, Number( 10 ) ) ), True )
  ] ) }

  a._test_vector = [
                'in_              *out',
    [        Bits16,           Bits16 ],
    [             0,               10 ],
    [             2,               12 ],
    [    Bits16(-1),        Bits16(9) ],
    [    Bits16(-2),        Bits16(8) ],
    [Bits16(0x7FFF),   Bits16(0x8009) ],
  ]

  do_test( a )

def test_index_bits_slicing( do_test ):
  class index_bits_slicing( Component ):
    def construct( s ):
      s.in_ = [ InPort( Bits16 ) for _ in range( 10 ) ]
      s.out = [ OutPort( Bits16 ) for _ in range( 5 ) ]

      @s.update
      def index_bits_slicing():
        s.out[0][0:8] = s.in_[1][8:16] + s.in_[2][0:8] + Bits8( 10 )
        s.out[1] = s.in_[3][0:16] + s.in_[4] + Bits16( 1 )

  a = index_bits_slicing()

  a._rtlir_test_ref = { 'index_bits_slicing' : CombUpblk( 'index_bits_slicing', [
    Assign( Slice(
      Index( Attribute( Base( a ), 'out' ), Number( 0 ) ),
      Number( 0 ), Number( 8 )
      ),
      BinOp(
        BinOp(
          Slice( Index( Attribute( Base( a ), 'in_' ), Number( 1 ) ), Number( 8 ), Number( 16 ) ),
          Add(),
          Slice( Index( Attribute( Base( a ), 'in_' ), Number( 2 ) ), Number( 0 ), Number( 8 ) ),
        ),
        Add(),
        SizeCast( 8, Number( 10 ) )
      ),
      True
    ),
    Assign(
      Index( Attribute( Base( a ), 'out' ), Number( 1 ) ),
      BinOp(
        BinOp(
          Slice( Index( Attribute( Base( a ), 'in_' ), Number( 3 ) ), Number( 0 ), Number( 16 ) ),
          Add(),
          Index( Attribute( Base( a ), 'in_' ), Number( 4 ) )
        ),
        Add(),
        SizeCast( 16, Number( 1 ) )
      ),
      True
    ),
  ] ) }

  a._test_vector = [
      'in_[0] in_[1] in_[2] in_[3] in_[4] in_[5] in_[6] in_[7] in_[8] in_[9]\
          *out[0] *out[1] *out[2] *out[3] *out[4]',
    [ Bits16 ] * 15,

    # 8-bit truncation!
    [ Bits16(0xff) ] * 10 + [ Bits16(0x09), Bits16(0x01ff), 0, 0, 0 ],
    [ Bits16(0x00) ] * 10 + [ Bits16(0x0a), Bits16(0x0001), 0, 0, 0 ],
  ]

  do_test( a )

def test_multi_components( do_test ):
  class multi_components_B( Component ):
    def construct( s ):
      s.in_ = InPort( Bits16 )
      s.out = OutPort( Bits16 )

      @s.update
      def multi_components_B():
        s.out = s.in_

  class multi_components_A( Component ):
    def construct( s ):
      s.in_ = InPort( Bits16 )
      s.out = OutPort( Bits16 )
      s.b = multi_components_B()

      # There should be a way to check module connections?
      connect( s.in_, s.b.in_ )

      @s.update
      def multi_components_A():
        s.out = s.in_ + s.b.out

  a = multi_components_A()

  a._rtlir_test_ref = { 'multi_components_A' : CombUpblk( 'multi_components_A', [
    Assign( Attribute( Base( a ), 'out' ),
      BinOp(
        Attribute( Base( a ), 'in_' ),
        Add(),
        Attribute( Attribute( Base( a ), 'b' ), 'out' )
      ),
      True
    )
  ] ) }

  a._test_vector = [
                'in_              *out',
    [ Bits16 ] *2,

    [             0,               0 ],
    [             2,               4 ],
    [    Bits16(-1),      Bits16(-2) ],
    [    Bits16(-2),      Bits16(-4) ],
  ]

  do_test( a )

def test_if_basic( do_test ):
  class if_basic( Component ):
    def construct( s ):
      s.in_ = InPort( Bits16 )
      s.out = OutPort( Bits8 )

      @s.update
      def if_basic():
        if s.in_[ 0:8 ] == Bits8( 255 ):
          s.out = s.in_[ 8:16 ]
        else:
          s.out = Bits8( 0 )

  a = if_basic()

  a._rtlir_test_ref = {
    'if_basic' : CombUpblk( 'if_basic', [ If(
      Compare( Slice( Attribute( Base( a ), 'in_' ), Number( 0 ), Number( 8 ) ), Eq(), SizeCast( 8, Number( 255 ) ) ),
      [ Assign( Attribute( Base( a ), 'out' ), Slice( Attribute( Base( a ), 'in_' ), Number( 8 ), Number( 16 ) ), True ) ],
      [ Assign( Attribute( Base( a ), 'out' ), SizeCast( 8, Number( 0 ) ), True ) ]
    )
  ] ) }

  a._test_vector = [
                'in_              *out',
    [         Bits16,           Bits8 ],
    [           255,                0 ],
    [           511,                1 ],
    [           256,                0 ],
  ]

  do_test( a )

@XFAIL_ON_PY3
def test_for_basic( do_test ):
  class for_basic( Component ):
    def construct( s ):
      s.in_ = InPort( Bits16 )
      s.out = OutPort( Bits8 )

      @s.update
      def for_basic():
        for i in range( 8 ):
          s.out[ 2*i:2*i+1 ] = s.in_[ 2*i:2*i+1 ] + s.in_[ 2*i+1:(2*i+1)+1 ]

  a = for_basic()

  twice_i = BinOp( Number( 2 ), Mult(), LoopVar( 'i' ) )

  a._rtlir_test_ref = {
    'for_basic' : CombUpblk( 'for_basic', [ For(
      LoopVarDecl( 'i' ), Number( 0 ), Number( 8 ), Number( 1 ),
      [ Assign(
          Slice( Attribute( Base( a ), 'out' ), twice_i, BinOp( twice_i, Add(), Number( 1 ) ) ),
          BinOp(
            Slice( Attribute( Base( a ), 'in_' ), twice_i, BinOp( twice_i, Add(), Number( 1 ) ) ),
            Add(),
            Slice( Attribute( Base( a ), 'in_' ),
              BinOp( twice_i, Add(), Number( 1 ) ),
              BinOp( BinOp( twice_i, Add(), Number( 1 ) ), Add(), Number( 1 ) )
            )
          ),
          True
        )
      ]
    ) ] )
  }

  a._test_vector = []

  do_test( a )

def test_multi_upblks( do_test ):
  class multi_upblks( Component ):
    def construct( s ):
      s.in_ = InPort( Bits4 )
      s.out = OutPort( Bits8 )

      @s.update
      def multi_upblks_1():
        s.out[ 0:4 ] = s.in_

      @s.update
      def multi_upblks_2():
        s.out[ 4:8 ] = s.in_

  a = multi_upblks()

  a._rtlir_test_ref = { 'multi_upblks_1' : CombUpblk( 'multi_upblks_1', [
      Assign( Slice( Attribute( Base( a ), 'out' ), Number(0), Number(4) ), Attribute( Base( a ), 'in_' ), True ),
    ] ),
    'multi_upblks_2' : CombUpblk( 'multi_upblks_2', [
      Assign( Slice( Attribute( Base( a ), 'out' ), Number(4), Number(8) ), Attribute( Base( a ), 'in_' ), True ),
    ] )
  }

  a._test_vector = [
                'in_              *out',
    [         Bits4,            Bits8 ],

    [     Bits4(-1),      Bits8(0xff) ],
    [      Bits4(1),      Bits8(0x11) ],
    [      Bits4(7),      Bits8(0x77) ],
  ]

  do_test( a )

def test_ff_upblk( do_test ):
  class A( Component ):
    def construct( s ):
      s.in0 = InPort( Bits32 )
      s.in1 = InPort( Bits32 )
      s.out0 = OutPort( Bits32 )
      @s.update_ff
      def update_ff_upblk():
        s.out0 <<= s.in0 + s.in1
  a = A()
  a._rtlir_test_ref = {
      'update_ff_upblk' : SeqUpblk( 'update_ff_upblk', [
        Assign( Attribute( Base( a ), 'out0' ), BinOp( Attribute( Base( a ), 'in0' ), Add(), Attribute( Base( a ), 'in1' ) ), False ),
        ] )
  }
  a._test_vector = [
                'in0                in1              *out',
    [         Bits32,            Bits32,            Bits32 ],

    [     Bits32(-1),       Bits32(0xff),     Bits32(0xfe) ],
    [     Bits32(1),        Bits32(0x11),     Bits32(0x12) ],
    [     Bits32(7),        Bits32(0x77),     Bits32(0x7d) ],
  ]

  do_test( a )

def test_fixed_size_slice( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( Bits16 )
      s.out = [ OutPort( Bits8 ) for _ in range(2) ]
      @s.update
      def upblk():
        for i in range(2):
          s.out[i] = s.in_[i*8 : i*8 + 8]
  a = A()
  a._rtlir_test_ref = {
      'upblk' : CombUpblk( 'upblk', [
        For( LoopVarDecl('i'), Number(0), Number(2), Number(1), [
          Assign(
            Index(Attribute(Base(a), 'out'), LoopVar('i')),
            Slice(Attribute(Base(a), 'in_'),
                  BinOp(LoopVar('i'), Mult(), Number(8)),
                  BinOp(BinOp(LoopVar('i'), Mult(), Number(8)), Add(), Number(8)),
                  BinOp(LoopVar('i'), Mult(), Number(8)),
                  8,
            ), True)]
        )
      ])
  }
  a._test_vector = [
                'in_              *out[0]          *out[1]',
    [         Bits16,               Bits8,            Bits8 ],

    [     Bits16(-1),         Bits8(0xff),      Bits8(0xff) ],
    [      Bits16(1),         Bits8(0x01),      Bits8(0x00) ],
    [      Bits16(7),         Bits8(0x07),      Bits8(0x00) ],
    [ Bits16(0xff00),         Bits8(0x00),      Bits8(0xff) ],
    [ Bits16(0x3412),         Bits8(0x12),      Bits8(0x34) ],
    [ Bits16(0x9876),         Bits8(0x76),      Bits8(0x98) ],
  ]

  do_test( a )
